# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2020, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Create an Analysis Period to describe a slice of time during the year.

-

    Args:
        _start_month_: Start month (1-12).
        _start_day_: Start day (1-31).
        _start_hour_: Start hour (0-23).
        _end_month_: End month (1-12).
        _end_day_: End day (1-31).
        _end_hour_: End hour (0-23).
        _timestep_: An integer number for the number of time steps per hours.
            Acceptable inputs include: 1, 2, 3, 4, 5, 6, 10, 12, 15, 20, 30, 60
    
    Returns:
        period: Analysis period.
        hoys: List of dates in this analysis period.
        dates: List of hours of the year in this analysis period.
"""

ghenv.Component.Name = 'LB Analysis Period'
ghenv.Component.NickName = 'AnalysisPeriod'
ghenv.Component.Message = '0.1.1'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

try:
    import ladybug.analysisperiod as ap
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import wrap_output
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


anp = ap.AnalysisPeriod(
    _start_month_, _start_day_, _start_hour_,
    _end_month_, _end_day_, _end_hour_, _timestep_)

if anp:
    period = anp
    dates = wrap_output(anp.datetimes)
    hoys = anp.hoys


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2020, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Output a Sunpath (aka. sun plot) graphic into the Rhino scene.
-
The component also outputs sun vectors that can be used for solar access
analysis and shading design.
-

    Args:
        north_: A number between -360 and 360 for the counterclockwise
            difference between the North and the positive Y-axis in degrees.
            90 is West and 270 is East. This can also be Vector for the
            direction to North. (Default: 0)
        _location: A ladybug Location that has been output from the "LB Import EPW"
            component or the "LB Construct Location" component.
        hoys_: A number or list of numbers between 0 and 8760 that respresent the
            hour of the year at which to evaluate the sun position. The
            "LB Calculate HOY" component can output this number given a month,
            day and hour. The "LB Analysis Period" component can output a
            list of HOYs within a certain hour or date range.
        dl_saving_: An optional analysis period for daylight saving time.
            If None, no daylight saving time will be used. (Default: None)
        solar_time_: A boolean to indicate if the input hours should be treated
             as solar time instead of standard or daylight time. (Default: False)
        _center_pt_: A point for the center of the sun path. (Default: (0, 0, 0))
        _scale_: A number to set the scale of the sun path. The default is 1,
            which corresponds to a radius of 100 meters in the current Rhino
            model's unit system.
        projection_: Optional text for the name of a projection to use from the sky
            dome hemisphere to the 2D plane. If None, a 3D sun path will be drawn
            instead of a 2D one. (Default: None) Choose from the following:
                * Orthographic
                * Stereographic
        daily_: Boolean to note whether the sunpath should display only one daily
            arc for each unique day in the input hoys_ (True) or whther the
            output sun path geometry should be for the entire year, complete
            with analemmas for all sun-up hours and a daily arc for each
            month (False). (Default: False)
        data_: Optional HourlyContinuousCollection objects, which will be used
            to generate colors that align with each of the sun_pts. This data
            can also be used along with the statement_ below to select out
            sun positions that meet certain conditions.
        statement_: A conditional statement as a string (e.g. a > 25).
            .   
            The variable of the first data collection should always be named 'a'
            (without quotations), the variable of the second list should be
            named 'b', and so on.
            .
            For example, if three data collections are connected to _data
            and the following statement is applied:
            '18 < a < 26 and b < 80 and c > 2'
            The resulting collections will only include values where the first
            data collection is between 18 and 26, the second collection is less
            than 80 and the third collection is greater than 2.
        legend_par_: An optional LegendParameter object to change the display
            of the data on the sun path. This can also be a list of legend
            parameters to be applied to the different connected data_.

    Returns:
        vectors: Vector(s) indicating the direction of sunlight for each sun
            position on the sun path. Note that these vectors point downwards
            towards the ground.
        altitudes: Number(s) indicating the sun altitude(s) in degrees for
            each sun position on the sun path.
        azimuths: Number(s) indicating the sun azimuths in degrees for each
            sun position on the sun path.
        hoys: The hour of the year for each sun positions on the sun path.
        sun_pts: Point(s) representing the location of the sun on the sunpath.
        analemma: A set of curves that mark the hourly positions of the sun
            throughout the different months of the year.
        daily: A set of arcs that mark the path of the sun across the sky
            dome over the course of a day.
        compass: A set of circles, lines and text objects that mark the cardinal
            directions in relation to the sun path.
        legend: Geometry representing the legend for the input data_. Will be None
            if no _data is connected.
        title: A text object for the title of the sunpath.
        color_pts: A list of points colored with the input data_, which will display
            in the Rhino scene in accordance with the legend. Note that
            the text representation of these objects bears the RGB color of each
            point. So casting this output to text and then to a color will
            yeild color objects that can be used for previewing other types
            of geometry with the input data_. Will be None if no _data is connected.
"""

ghenv.Component.Name = 'LB SunPath'
ghenv.Component.NickName = 'Sunpath'
ghenv.Component.Message = '0.2.5'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '2 :: Visualize Data'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

try:
    from ladybug_geometry.geometry2d.pointvector import Vector2D, Point2D
    from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
    from ladybug_geometry.geometry3d.plane import Plane
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug.sunpath import Sunpath
    from ladybug.compass import Compass
    from ladybug.graphic import GraphicContainer
    from ladybug.datacollection import HourlyContinuousCollection
    from ladybug.dt import Date
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import conversion_to_meters
    from ladybug_rhino.color import color_to_color
    from ladybug_rhino.colorize import ColoredPoint
    from ladybug_rhino.fromgeometry import from_polyline3d, from_polyline2d, \
        from_arc3d, from_vector3d, from_point3d, from_point2d
    from ladybug_rhino.fromobjects import legend_objects, compass_objects
    from ladybug_rhino.togeometry import to_vector2d, to_point2d, to_point3d
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.grasshopper import all_required_inputs, list_to_data_tree, \
        hide_output, show_output, schedule_solution
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

import math


def draw_analemma_and_arcs(sp, datetimes, radius, center_pt3d):
    """Draw analemma and day arc Rhino geometry.

    Args:
        sp: Sunpath object for which geometry will be drawn.
        datetimes: A list of datetimes, which will be used to get days
            if daily_ is True.
        radius: Number for the radius of the sun path.
        center_pt3d: Point3D for the center of the sun path.

    Returns:
        analemma: List of Rhino curves for the analemmas
        daily: List of Rhino curves for the daily arcs.
    """
    sp.daylight_saving_period = None  # set here so analemmas aren't messed up

    center_pt, z = Point2D(center_pt3d.x, center_pt3d.y), center_pt3d.z
    if not daily_:
        if projection_ is None:
            analemma = [from_polyline3d(pline) for pline in sp.hourly_analemma_polyline3d(
                center_pt3d, radius, True, solar_time_)]
            daily = [from_arc3d(arc) for arc in sp.monthly_day_arc3d(center_pt3d, radius)]
        else:
            analemma = [from_polyline2d(pline, z) for pline in sp.hourly_analemma_polyline2d(
                projection_, center_pt, radius, True, solar_time_)]
            daily = [from_polyline2d(arc, z) for arc in sp.monthly_day_polyline2d(
                projection_, center_pt3d, radius)]
    else:
        analemma = []  # No Analemmas for a daily sun path
        doys = set(dt.doy for dt in datetimes)
        dates = [Date.from_doy(doy) for doy in doys]
        if projection_ is None:
            daily = [from_arc3d(sp.day_arc3d(dat.month, dat.day, center_pt3d, radius))
                     for dat in dates]
        else:
            daily = []
            for dat in dates:
                pline = sp.day_polyline2d(dat.month, dat.day, projection_, center_pt, radius)
                daily.append(from_polyline2d(pline, z))
    return analemma, daily


def draw_sun_positions(suns, radius, center_pt3d):
    """Draw Rhino points from a list of sun objects.

    Args:
        suns: A list of suns to be output as points
        radius: Number for the radius of the sun path.
        center_pt3d: Point3D for the center of the sun path.

    Returns:
        A list of Rhino points for sun positions
    """
        # get points for sun positions
    if projection_ is None:
        return [from_point3d(sun.position_3d(center_pt3d, radius)) for sun in suns]
    else:
        return [from_point2d(sun.position_2d(projection_, center_pt3d, radius), z)
                for sun in suns]


def title_text(data_col):
    """Get a text string for the title of the sunpath."""
    title_array = ['{} ({})'.format(data_col.header.data_type,
                                    data_col.header.unit)]
    for key, val in data_col.header.metadata.items():
        title_array.append('{}: {}'.format(key, val))
    return '\n'.join(title_array)


if all_required_inputs(ghenv.Component):
    # process all of the global inputs for the sunpath
    if north_ is not None:  # process the north_
        try:
            north_ = math.degrees(
                to_vector2d(north_).angle_clockwise(Vector2D(0, 1)))
        except AttributeError:  # north angle instead of vector
            north_ = float(north_)
    else:
        north_ = 0
    if _center_pt_ is not None:  # process the center point into a Point2D
        center_pt, center_pt3d = to_point2d(_center_pt_), to_point3d(_center_pt_)
        z = center_pt3d.z
    else:
        center_pt, center_pt3d = Point2D(), Point3D()
        z = 0
    _scale_ = 1 if _scale_ is None else _scale_ # process the scale into a radius
    radius = (100 * _scale_) / conversion_to_meters()
    solar_time_ = False if solar_time_ is None else solar_time_  # process solat time

    # create a intersection of the input hoys_ and the data hoys
    if len(data_) > 0 and len(hoys_) > 0:
        all_aligned = all(data_[0].is_collection_aligned(d) for d in data_[1:])
        assert all_aligned, 'All collections input to data_ must be aligned for ' \
            'each Sunpath.\nGrafting the data_ and suplying multiple grafted ' \
            '_center_pt_ can be used to view each data on its own path.'
        if statement_ is not None:
            data_ = HourlyContinuousCollection.filter_collections_by_statement(
                data_, statement_)
        data_hoys = set(dt.hoy for dt in data_[0].datetimes)
        hoys_ = list(data_hoys.intersection(set(hoys_)))

    # initialize sunpath based on location
    sp = Sunpath.from_location(_location, north_, dl_saving_)

    # process all of the input hoys into altitudes, azimuths and vectors
    altitudes, azimuths, datetimes, moys, hoys, vectors, suns = [], [], [], [], [], [], []
    for hoy in hoys_:
        sun = sp.calculate_sun_from_hoy(hoy, solar_time_)
        if sun.is_during_day:
            altitudes.append(sun.altitude)
            azimuths.append(sun.azimuth)
            datetimes.append(sun.datetime)
            moys.append(sun.datetime.moy)
            hoys.append(sun.datetime.hoy)
            vectors.append(from_vector3d(sun.sun_vector))
            suns.append(sun)

    if len(data_) > 0 and len(hoys_) > 0:  # build a sunpath for each data collection
        title, all_sun_pts, all_analemma, all_daily, all_compass, all_col_pts, all_legends = \
            [], [], [], [], [], [], []
        for i, data in enumerate(data_):
            try:  # sense when several legend parameters are connected
                lpar = legend_par_[i]
            except IndexError:
                lpar = None if len(legend_par_) == 0 else legend_par_[-1]

            # move the center point so sun paths are not on top of one another
            fac = i* radius * 3
            center_pt_i = Point2D(center_pt.x + fac, center_pt.y)
            center_pt3d_i = Point3D(center_pt3d.x + fac, center_pt3d.y, center_pt3d.z)

            # create the ladybug compass object
            lb_compass = Compass(radius, center_pt_i, north_)

            # create a graphic container to generate colors and legends
            n_data = data.filter_by_moys(moys)  # filter data collection by sun-up hours
            graphic = GraphicContainer(
                n_data.values, lb_compass.min_point3d(z), lb_compass.max_point3d(z),
                lpar, n_data.header.data_type, n_data.header.unit)
            all_legends.append(legend_objects(graphic.legend))
            title.append(text_objects(
                title_text(n_data), graphic.lower_title_location,
                graphic.legend_parameters.text_height, graphic.legend_parameters.font))

            # create points, analemmas, daily arcs, and compass geometry
            sun_pts_init = draw_sun_positions(suns, radius, center_pt3d_i)
            analemma_i, daily_i = draw_analemma_and_arcs(sp, datetimes, radius, center_pt3d_i)
            compass_i = compass_objects(lb_compass, z, None, projection_,
                                        graphic.legend_parameters.font)
            all_analemma.append(analemma_i)
            all_daily.append(daily_i)
            all_compass.append(compass_i)

            # produce a visualization of colored points
            cols = [color_to_color(col) for col in graphic.value_colors]
            col_pts = []
            for pt, col in zip(sun_pts_init, cols):
                col_pt = ColoredPoint(pt)
                col_pt.color = col
                col_pts.append(col_pt)
            all_sun_pts.append(sun_pts_init)
            all_col_pts.append(col_pts)

        # convert all nested lists to data trees
        sun_pts = list_to_data_tree(all_sun_pts)
        analemma = list_to_data_tree(all_analemma)
        daily = list_to_data_tree(all_daily)
        compass = list_to_data_tree(all_compass)
        legend = list_to_data_tree(all_legends)

        # do some acrobatics to get the colored points to display
        # CWM: I don't know why we have to re-schedule the solution but this is the
        # only way I found to get the colored points to appear (redraw did not work).
        color_pts = list_to_data_tree(all_col_pts)
        hide_output(ghenv.Component, 5)
        schedule_solution(ghenv.Component, 2)
    else:  # no data connected; just output one sunpath
        sun_pts = draw_sun_positions(suns, radius, center_pt3d)
        analemma, daily = draw_analemma_and_arcs(sp, datetimes, radius, center_pt3d)
        font = legend_par_[0].font if len(legend_par_) != 0 and \
            legend_par_[0] is not None else 'Arial'
        compass = compass_objects(Compass(radius, center_pt, north_), z, None, projection_, font)
        if _location.city:
            title = text_objects(
                'city: {}'.format(_location.city),
                Plane(o=center_pt3d.move(Vector3D(-radius * 1.25, -radius * 1.25))),
                radius / 15, font)
        show_output(ghenv.Component, 5)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2020, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Get a matrix containing radiation values from each patch of a sky dome.
_
Creating this matrix is a necessary pre-step before doing incident radiation
analysis with Rhino geometry or generating a radiation rose.
_
This component uses Radiance's gendaymtx function to calculate the radiation
for each patch of the sky. Gendaymtx is written by Ian Ashdown and Greg Ward.
Morere information can be found in Radiance manual at:
http://www.radiance-online.org/learning/documentation/manual-pages/pdfs/gendaymtx.pdf
-

    Args:
        north_: A number between -360 and 360 for the counterclockwise
            difference between the North and the positive Y-axis in degrees.
            90 is West and 270 is East. This can also be Vector for the
            direction to North. (Default: 0)
        _location: A ladybug Location that has been output from the "LB Import EPW"
            component or the "LB Construct Location" component.
        _direct_rad: An annual hourly DataCollection of Direct Normal Radiation such
            as that which is output from the "LB Import EPW" component or the
            "LB Import STAT" component.
        _diffuse_rad: An annual hourly DataCollection of Diffuse Horizontal Radiation
            such as that which is output from the "LB Import EPW" component or
            the "LB Import STAT" component.
        _hoys_: A number or list of numbers between 0 and 8760 that respresent
            the hour(s) of the year for which to generate the sky matrix. The
            "LB Calculate HOY" component can output this number given a month,
            day and hour. The "LB Analysis Period" component can output a
            list of HOYs within a certain hour or date range. By default,
            the matrix will be for the entire year.
        high_density_: A Boolean to indicate whether the higher-density Reinhart
            sky matrix should be generated (True), which has roughly 4 times
            the sky patches as the (default) original Tregenza sky (False).
            Note that, while the Reinhart sky has a higher resolution and is
            more accurate, it will result in considerably longer calculation
            time for incident radiation studies. The difference in sky
            resolution can be observed with the (Default: False).
        _folder_: The folder in which the Radiance commands are executed to
            produce the sky matrix. If None, it will be written to Ladybug's
            default EPW folder.

    Returns:
        report: ...
        sky_mtx: A sky matrix object containing the radiation coming from each patch
            of the sky. This can be used for a radiation study, a radition rose,
            or a sky dome visualization. It can also be deconstructed into its
            individual values with the "LB Deconstruct Matrix" component.
"""

ghenv.Component.Name = 'LB Cumulative Sky Matrix'
ghenv.Component.NickName = 'SkyMatrix'
ghenv.Component.Message = '0.1.3'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '2 :: Visualize Data'
ghenv.Component.AdditionalHelpFromDocStrings = '3'

import os
import subprocess
import math

try:
    from ladybug_geometry.geometry2d.pointvector import Vector2D
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug.wea import Wea
    from ladybug.viewsphere import view_sphere
    from ladybug.dt import DateTime
    from ladybug.config import folders as lb_folders
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.togeometry import to_vector2d
    from ladybug_rhino.grasshopper import all_required_inputs, objectify_output
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


# TODO: Remove dependency on honeybee + Radiance after genskymtx is in its own LB extension
try:
    from honeybee_radiance.config import folders as hb_folders
    rad_url = 'https://github.com/LBNL-ETA/Radiance/releases/tag/012cb178'
    assert hb_folders.radiance_path is not None, \
        'No Radiance installation was found on this machine.\n' \
        'Download and install Radiance from\n{}.'.format(rad_url)
    assert hb_folders.radiance_version[:2] >= (5, 3), \
        'The installed Radiance is not version 5.3 or greater.\n' \
        'Download and install Radiance from\n{}.'.format(rad_url)
    # get the path to the gemdaymtx executable
    gendaymtx_exe = os.path.join(hb_folders.radbin_path, 'gendaymtx.exe') if \
        os.name == 'nt' else os.path.join(hb_folders.radbin_path, 'gendaymtx')
except ImportError as e:
    raise ImportError('\nFailed to import honeybee_radiance:\n\t{}'.format(e))


# constants for converting RGB values output by gendaymtx to broadband radiation
PATCHES_PER_ROW = {
    1: view_sphere.TREGENZA_PATCHES_PER_ROW + (1,),
    2: view_sphere.REINHART_PATCHES_PER_ROW + (1,)
}
PATCH_ROW_COEFF = {
    1 : view_sphere.TREGENZA_COEFFICIENTS,
    2: view_sphere.REINHART_COEFFICIENTS
}


def broadband_radiation(patch_row_str, row_number, wea_duration, sky_density=1):
    """Parse a row of gendaymtx RGB patch data in W/sr/m2 to radiation in kWh/m2.

    This includes aplying broadband weighting to the RGB bands, multiplication
    by the steradians of each patch, and multiplying by the duration of time that
    they sky matrix represents in hours.

    Args:
        patch_row_str: Text string for a single row of RGB patch data.
        row_number: Interger for the row number that the patch corresponds to.
        sky_density: Integer (either 1 or 2) for the density.
        wea_duration: Number for the duration of the Wea in hours. This is used
            to convert between the average value output by the command and the
            cumulative value that is needed for all ladybug analyses.
    """
    R, G, B = patch_row_str.split(' ')
    weight_val = 0.265074126 * float(R) + 0.670114631 * float(G) + 0.064811243 * float(B)
    return weight_val * PATCH_ROW_COEFF[sky_density][row_number] * wea_duration / 1000


def parse_mtx_data(data_str, wea_duration, sky_density=1):
    """Parse a string of Radiance gendaymtx data to a list of radiation-per-patch.

    This function handles the removing of the header and the conversion of the
    RGB irradianc-=per-steraidian values to broadband radiation. It also removes
    the first patch, which is the ground and is not used by Ladybug.

    Args:
        data_str: The string that has been output by gendaymtx to stdout.
        wea_duration: Number for the duration of the Wea in hours. This is used
            to convert between the average value output by the command and the
            cumulative value that is needed for all ladybug analyses.
        sky_density: Integer (either 1 or 2) for the density.
    """
    # split lines and remove the header, ground patch and last line break
    data_lines = data_str.split('\n')
    patch_lines = data_lines[9:-1]

    # loop through the rows and convert the radiation RGB values
    broadband_irr = []
    patch_counter = 0
    for i, row_patch_count in enumerate(PATCHES_PER_ROW[sky_density]):
        row_slice = patch_lines[patch_counter:patch_counter + row_patch_count]
        irr_vals = (broadband_radiation(row, i, wea_duration, sky_density)
                    for row in row_slice)
        broadband_irr.extend(irr_vals)
        patch_counter += row_patch_count
    return broadband_irr


if all_required_inputs(ghenv.Component):
    # process and set defaults for all of the global inputs
    if north_ is not None:  # process the north_
        try:
            north_ = math.degrees(
                to_vector2d(north_).angle_clockwise(Vector2D(0, 1)))
        except AttributeError:  # north angle instead of vector
            north_ = float(north_)
    else:
        north_ = 0
    density = 2 if high_density_ else 1

    # filter the radiation by _hoys if they are input
    if len(_hoys_) != 0:
        _direct_rad = _direct_rad.filter_by_hoys(_hoys_)
        _diffuse_rad = _diffuse_rad.filter_by_hoys(_hoys_)

    # create the wea and write it to the default_epw_folder
    wea = Wea(_location, _direct_rad, _diffuse_rad)
    wea_duration = len(wea) / wea.timestep
    wea_folder = _folder_ if _folder_ is not None else \
        os.path.join(lb_folders.default_epw_folder, 'sky_matrices')
    metd = _direct_rad.header.metadata
    wea_basename = os.path.join(wea_folder, metd['city'].replace(' ', '_')) \
        if 'city' in metd else 'unnamed'
    wea_file =wea.write(wea_basename)

    # execute the Radiance gendaymtx command
    use_shell = True if os.name == 'nt' else False
    # command for direct patches
    cmds = [gendaymtx_exe, '-m', str(density), '-d', '-O1', '-A', wea_file]
    process = subprocess.Popen(cmds, stdout=subprocess.PIPE, shell=use_shell)
    stdout = process.communicate()
    dir_data_str = stdout[0]
    # command for diffuse patches
    cmds = [gendaymtx_exe, '-m', str(density), '-s', '-O1', '-A', wea_file]
    process = subprocess.Popen(cmds, stdout=subprocess.PIPE, shell=use_shell)
    stdout = process.communicate()
    diff_data_str = stdout[0]

    # parse the data into a single matrix
    dir_vals = parse_mtx_data(dir_data_str, wea_duration, density)
    diff_vals = parse_mtx_data(diff_data_str, wea_duration, density)

    # collect sky metadata like the north, which will be used by other components
    metadata = [north_]
    if _hoys_:
        metadata.extend([DateTime.from_hoy(h) for h in (_hoys_[0], _hoys_[-1])])
    else:
        metadata.extend([wea.analysis_period.st_time, wea.analysis_period.end_time])
    for key, val in _direct_rad.header.metadata.items():
        metadata.append('{} : {}'.format(key, val))

    # wrap everything together into an object to output from the component
    mtx_data = (metadata, dir_vals, diff_vals)
    sky_mtx = objectify_output('Cumulative Sky Matrix', mtx_data)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2020, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Genrate a mesh with corresponding test points.
The resulting mesh will be in a format that the 'Color Mesh' component will accept.
-

    Args:
        _geometry: Brep or Mesh from which to generate the points and grid.
        _grid_size: Number for the size of the test grid.
        _offset_dist_: Number for the distance to move points from the surfaces
            of the input _geometry.  Typically, this should be a small positive
            number to ensure points are not blocked by the mesh. Default is 0.
    
    Returns:
        points: Test points at the center of each mesh face.
        vectors: Vectors for the normal direction at each of the points.
        face_areas: Area of each mesh face.
        mesh: Analysis mesh that can be passed to the 'Color Mesh' component.
"""

ghenv.Component.Name = "LB Generate Point Grid"
ghenv.Component.NickName = 'GenPts'
ghenv.Component.Message = '0.1.4'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from ladybug_rhino.togeometry import to_gridded_mesh3d, to_mesh3d
    from ladybug_rhino.fromgeometry import from_mesh3d, from_point3d, from_vector3d
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # check the input and generate the mesh.
    _offset_dist_ = _offset_dist_ or 0
    try:  # assume it's a Rhino Brep
        lb_mesh = to_gridded_mesh3d(_geometry, _grid_size, _offset_dist_)
    except TypeError:  # assume it's a Rhino Mesh
        try:
            lb_mesh = to_mesh3d(_geometry)
        except TypeError:  # unidientified geometry type
            raise TypeError(
                '_geometry must be a Brep or a Mesh. Got {}.'.format(type(_geometry)))
    
    # generate the test points, vectors, and areas.
    points = [from_point3d(pt) for pt in lb_mesh.face_centroids]
    vectors = [from_vector3d(vec) for vec in lb_mesh.face_normals]
    face_areas = lb_mesh.face_areas
    mesh = from_mesh3d(lb_mesh)


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2020, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Evaluate the percent view to the outdoors or sky from input geometry through context.
_
Such view calculations can be used to estimate the quality of a view to the
outdoors from a given location on the indoors. They can also be used on
the outdoors to evaluate the openness of street canyons to the sky, which has
implications for the pedestrian expereince as well as the rate of radiant heat
loss from urban surfaces and the sky at night.
_
Note that this component uses the CAD environment's ray intersection methods,
which can be fast for geometries with low complexity but does not scale well
for complex geometries or many test points. For such complex studies,
honeybee-radiance should be used.
-

    Args:
        _view_type: Text or an integer representing the type of view analysis
            to conduct.  Choose from the following options.
                _
                0 - HorizontalRadial - The percentage of the 360 horizontal view
                    plane that is not blocked by the context geometry.
                _
                1 - Horizonta30DegreeOffset - The percentage of the 360 horizontal
                    view band bounded on top and bottom by a 30 degree offset from
                    the horizontal plane. 30 degress corresponds roughly to the
                    vertical limit of human peripheral vision.
                _
                2 - Spherical - The percentage of the sphere surrounding each of
                    the test points that is not blocked by context geometry. This
                    is equivalent to a solid angle and gives equal weight to all
                    portions of the sphere.
                _
                3 - SkyExposure - The percentage of the sky that is visible from
                    each of the the test points. This is distinct from SkyView,
                    which is the amount of sky seen by a surface. SkyExposure is
                    equivalent to a solid angle and gives equal weight to all
                    portions of the sky.
                _
                4 - SkyView - The percentage of the sky that is visible from the
                    _geometry surfaces. This is distinct from SkyExposure, which
                    treats each part of the sky with equal weight. SkyView weights
                    the portions of the sky according to thier projection into
                    the plane of the surface being evaluated. So SkyView for a
                    horizontal surface would give more importance to the sky
                    patches that are overhead vs. those near the horizon.
        _resolution_: A positive integer for the number of times that the original
            view vectors are subdivided. 1 indicates that 145 evenly-spaced
            vectors are used to describe a hemisphere, 2 indicates that
            577 vectors describe a hemisphere, and each successive value will
            roughly quadruple the number of view vectors used. Setting this to
            a high value will result in a more accurate analysis but will take
            longer to run. (Default: 1).
        _geometry: Rhino Breps and/or Rhino Meshes for which view analysis
            will be conducted. If Breps are input, they will be subdivided using
            the _grid_size to yeild individual points at which analysis will
            occur. If a Mesh is input, view analysis analysis will be
            performed for each face of this mesh instead of subdividing it.
        context_: Rhino Breps and/or Rhino Meshes representing context geometry
            that can block view from the test _geometry.
        _grid_size: A positive number in Rhino model units for the size of grid
            cells at which the input _geometry will be subdivided for direct sun
            analysis. The smaller the grid size, the higher the resolution of
            the analysis and the longer the calculation will take.  So it is
            recommended that one start with a large value here and decrease
            the value as needed. However, the grid size should usually be
            smaller than the dimensions of the smallest piece of the _geometry
            and context_ in order to yield meaningful results.
        _offset_dist_: A number for the distance to move points from the surfaces
            of the input _geometry.  Typically, this should be a small positive
            number to ensure points are not blocked by the mesh. (Default: 10 cm
            in the equivalent Rhino Model units).
        _geo_block_: Set to "True" to count the input _geometry as opaque and
            set to "False" to discount the _geometry from the calculation and
            only look at context_ that blocks the view.  The default depends on
            the _view_type used.
            _
            It is "True" for:
            * SkyExposure
            * SkyView
            _
            It is "False" for:
            * HorizontalRadial
            * Horizonta30DegreeOffset
            * Spherical
        legend_par_: Optional legend parameters from the "LB Legend Parameters"
            that will be used to customize the display of the results.
        parallel_: Set to "True" to run the study using multiple CPUs. This can
            dramatically decrease calculation time but can interfere with
            other computational processes that might be running on your
            machine. (Default: False).
        _run: Set to "True" to run the component and perform view analysis of
            the input _geometry.

    Returns:
        report: ...
        points: The grid of points on the test _geometry that are be used to perform
            the view analysis.  Note that these points are generated even
            when _run is set to "False" so that an estimate of study run time can
            be obtained.
        view_vecs: A list of vectors which are projected from each of the points
            to evaluate view. Note that these vectors are generated even
            when _run is set to "False" so that an understanding of the study's
            field fo view and resolution can be obtained before running the study.
        results: A list of numbers that aligns with the points. Each number indicates
            the percentage of the view_vecs that are not blocked by context geometry.
        mesh: A colored mesh of the test _geometry representing the percentage of
            the input _geometry's view that is not blocked by context.
        legend: A legend showing the number of hours that correspond to the colors
            of the mesh.
        title: A text object for the study title.
        int_mtx: A Matrix object that can be connected to the "LB Deconstruct Matrix"
            component to obtain detailed vector-by-vector results of the study.
            Each sub-list (aka. branch of the Data Tree) represents one of the
            points used for analysis. The length of each sub-list matches the
            number of view_vecs used for the analysis. Each value in the sub-list
            is either a "1", indicating that the vector is visible for that
            vector, or a "0", indicating that the vector is not visible for
            that vector.
"""

ghenv.Component.Name = "LB View Percent"
ghenv.Component.NickName = 'ViewPercent'
ghenv.Component.Message = '0.1.2'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '3 :: Analyze Geometry'
ghenv.Component.AdditionalHelpFromDocStrings = '2'

import math
try:  # python 2
    from itertools import izip as zip
except ImportError:  # python 3
    pass

try:
    from ladybug.viewsphere import view_sphere
    from ladybug.color import Colorset
    from ladybug.graphic import GraphicContainer
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import conversion_to_meters
    from ladybug_rhino.togeometry import to_joined_gridded_mesh3d, to_vector3d
    from ladybug_rhino.fromgeometry import from_mesh3d, from_point3d, from_vector3d
    from ladybug_rhino.fromobjects import legend_objects
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.intersect import join_geometry_to_mesh, intersect_mesh_rays
    from ladybug_rhino.grasshopper import all_required_inputs, hide_output, \
        show_output, objectify_output
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


# dictionary to record all available view types
VIEW_TYPES = {
    'HorizontalRadial': 'Horizontal Radial',
    'Horizonta30DegreeOffset': 'Horizontal 30-Degree Offset',
    'Spherical': 'Spherical',
    'SkyExposure': 'Sky Exposure',
    'SkyView': 'Sky View',
    '0': 'Horizontal Radial',
    '1': 'Horizontal 30-Degree Offset',
    '2': 'Spherical',
    '3': 'Sky Exposure',
    '4': 'Sky View'
}


if all_required_inputs(ghenv.Component):
    # process the view_type_ and set the default values
    vt_str = VIEW_TYPES[_view_type]
    _resolution_ = _resolution_ if _resolution_ is not None else 1
    _offset_dist_ = _offset_dist_ if _offset_dist_ is not None \
        else 0.1 / conversion_to_meters()
    if _geo_block_ is None:
        _geo_block_ = True if vt_str in ('Sky Exposure', 'Sky View') else False

    # create the gridded mesh from the geometry
    study_mesh = to_joined_gridded_mesh3d(_geometry, _grid_size, _offset_dist_)
    points = [from_point3d(pt) for pt in study_mesh.face_centroids]

    # get the view vectors based on the view type
    patch_wghts = None
    if vt_str == 'Horizontal Radial':
        lb_vecs = view_sphere.horizontal_radial_vectors(30 * _resolution_)
    elif vt_str == 'Horizontal 30-Degree Offset':
        patch_mesh, lb_vecs = view_sphere.horizontal_radial_patches(30, _resolution_)
        patch_wghts = view_sphere.horizontal_radial_patch_weights(30, _resolution_)
    elif vt_str == 'Spherical':
        patch_mesh, lb_vecs = view_sphere.sphere_patches(_resolution_)
        patch_wghts = view_sphere.sphere_patch_weights(_resolution_)
    else:
        patch_mesh, lb_vecs = view_sphere.dome_patches(_resolution_)
        patch_wghts = view_sphere.dome_patch_weights(_resolution_)
    view_vecs = [from_vector3d(pt) for pt in lb_vecs]

    if _run:  # run the entire study
        # hide the study points and mesh the geometry and context
        hide_output(ghenv.Component, 1)
        shade_mesh = join_geometry_to_mesh(_geometry + context_) if _geo_block_ \
            else join_geometry_to_mesh(context_)

        # intersect the rays with the mesh
        if vt_str == 'Sky View':  # account for the normals of the surface
            normals = [from_vector3d(vec) for vec in study_mesh.face_normals]
            int_matrix, angles = intersect_mesh_rays(
                shade_mesh, points, view_vecs, normals, parallel=parallel_)
        else:
            int_matrix, angles = intersect_mesh_rays(
                shade_mesh, points, view_vecs, parallel=parallel_)

        # compute the results
        int_mtx = objectify_output('View Intersection Matrix', int_matrix)
        vec_count = len(view_vecs)
        results = []
        if vt_str == 'Sky View':  # weight intersections by angle before output
            for int_vals, angles in zip(int_matrix, angles):
                w_res = (ival * 2 * math.cos(ang) for ival, ang in zip(int_vals, angles))
                weight_result = sum(r * w for r, w in zip(w_res, patch_wghts))
                results.append(weight_result * 100 / vec_count)
        else:
            if patch_wghts:
                for int_list in int_matrix:
                    weight_result = sum(r * w for r, w in zip(int_list, patch_wghts))
                    results.append(weight_result * 100 / vec_count)
            else:
                results = [sum(int_list) * 100 / vec_count for int_list in int_matrix]

        # create the mesh and legend outputs
        graphic = GraphicContainer(results, study_mesh.min, study_mesh.max, legend_par_)
        graphic.legend_parameters.title = '%'
        if legend_par_ is None or legend_par_.are_colors_default:
            graphic.legend_parameters.colors = Colorset.view_study()
        title_txt = vt_str if vt_str in ('Sky Exposure', 'Sky View') else \
            '{} View'.format(vt_str)
        title = text_objects(
            title_txt, graphic.lower_title_location,
            graphic.legend_parameters.text_height * 1.5,
            graphic.legend_parameters.font)

        # create all of the visual outputs
        study_mesh.colors = graphic.value_colors
        mesh = from_mesh3d(study_mesh)
        legend = legend_objects(graphic.legend)

    else:  # only show the points and the view vectors
        show_output(ghenv.Component, 1)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2020, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Import hourly climate data from a Ladybug DesignDay object.
-

    Args:
        _design_day: A DesignDay object to import.
    
    Returns:
        location: A Ladybug Location object describing the location of the
            design day.
        dry_bulb_temperature: The houlry dry bulb temperature over the design
            day, in C.
        dew_point_temperature: The hourly dew point temperature over the design
            day, in C.
        relative_humidity: The hourly Relative Humidity over the design day
            in percent.
        wind_speed: The hourly wind speed over the design day in m/sec.
        wind_direction: The hourly wind direction over the design day in degrees.
            The convention is that North=0.0, East=90.0, South=180.0, West=270.0.
        direct_normal_rad: The hourly Direct Normal Radiation over the design
            day in Wh/m2.
        diffuse_horizontal_rad: The hourly Diffuse Horizontal Radiation over
            the design day in Wh/m2.
        global_horizontal_rad: The hourly Global Horizontal Radiation over the
            design day in Wh/m2.
        horizontal_infrared_rad: The Horizontal Infrared Radiation Intensity
            over the design day in Wh/m2.
        total_sky_cover: The fraction for total sky cover over the design day
            in tenths of coverage.
        barometric_pressure: The hourly weather station pressure over the
            design day in Pa.
"""

ghenv.Component.Name = 'LB Import Design Day'
ghenv.Component.NickName = 'ImportDesignDay'
ghenv.Component.Message = '0.1.1'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '0 :: Import'
ghenv.Component.AdditionalHelpFromDocStrings = '5'

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    dry_bulb_temperature = _design_day.hourly_dry_bulb
    dew_point_temperature = _design_day.hourly_dew_point
    relative_humidity = _design_day.hourly_relative_humidity
    wind_speed = _design_day.hourly_wind_speed
    wind_direction = _design_day.hourly_wind_direction
    direct_normal_rad, diffuse_horizontal_rad, global_horizontal_rad = \
        _design_day.hourly_solar_radiation
    horizontal_infrared_rad = _design_day.hourly_horizontal_infrared
    total_sky_cover = _design_day.hourly_sky_cover
    barometric_pressure = _design_day.hourly_barometric_pressure



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2020, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Calculate Universal Thermal Climate Index (UTCI).
-
UTCI is a thermal comfort model strictly for the outdoors.
It is an interational standard for outdoor temperature sensation
(aka. "feels-like" temperature) and is the most common of such
"feels-like" temperature metrics used by meteorologists.
-
While UTCI is valid in all climates, seasons, and scales, it assumes
that human subjects are walking and that they naturally adapt their
clothing with the outdoor temperature.
For outdoor situations that do not fit these criteria, the Physiological
Equivalent Temperature (PET) model should be used.

-

    Args:
        _air_temp: Data Collection or individual value of air temperature in C.
        _mrt_: Data Collection or individual value of mean radiant temperature
            (MRT) in C. Default is the same as the air_temp.
        _rel_humid: Data Collection or individual value of relative humidity in %.
        _wind_vel_: Data Collection or individual of air speed values in m/s.
            Default is a low speed of 0.5 m/s, which is the lowest input
            speed that is recommended for the UTCI model.
        _run: Set to True to run the component.

    Returns:
        report: Reports, errors, warnings, etc.
        utci: Universal Thermal Climate Index (UTCI) in Celcius.
        comfort: Integers noting whether the input conditions result in no
            thermal stress.
            .
            Values are one of the following:
                0 = thermal stress
                1 = no thremal stress
        condition: Integers noting the thermal status of a subject.
            .
            Values are one of the following:
                -1 = cold
                 0 = netural
                +1 = hot
        category: Integers noting the category that the UTCI conditions fall
            under on an 11-point scale.
            .
            Values are one of the following:
                -5 = Extreme Cold Stress       (UTCI < -40)
                -4 = Very Strong Cold Stress   (-40 <= UTCI < -27)
                -3 = Strong Cold Stress        (-27 <= UTCI < -13)
                -2 = Moderate Cold Stress      (-12 <= UTCI < 0)
                -1 = Slight Cold Stress        (0 <= UTCI < 9)
                 0 = No Thermal Stress         (9 <= UTCI < 26)
                +1 = Slight Heat Stress        (26 <= UTCI < 28)
                +2 = Moderate Heat Stress      (28 <= UTCI < 32)
                +3 = Strong Heat Stress        (32 <= UTCI < 38)
                +4 = Very Strong Heat Stress   (38 <= UTCI < 46)
                +5 = Extreme Heat Stress       (46 < UTCI)
        comf_obj: A Python object containing all inputs and results of the
            analysis.  This can be plugged into components like the "Comfort
            Statistics" component to get further information.
"""

ghenv.Component.Name = 'LB UTCI Comfort'
ghenv.Component.NickName = 'UTCI'
ghenv.Component.Message = '0.1.5'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '5'

try:
    from ladybug.datatype.temperature import Temperature
    from ladybug.datacollection import BaseCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_comfort.collection.utci import UTCI
    from ladybug_comfort.parameter.utci import UTCIParameter
    from ladybug_comfort.utci import universal_thermal_climate_index
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_comfort:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def extract_collections(input_list):
    """Process inputs into collections and floats."""
    defaults = [None, _air_temp, 0.5, None]
    data_colls = []
    for i, input in enumerate(input_list):
        if input is None:
            input_list[i] = defaults[i]
        elif isinstance(input, BaseCollection):
            data_colls.append(input)
        else:
            try:
                input_list[i] = float(input)
            except ValueError as e:
                raise TypeError('input {} is not valid. Expected float or '
                                'DataCollection. Got {}'.format(input, type(input)))
    return input_list, data_colls

if all_required_inputs(ghenv.Component) and _run is True:
    # Process inputs and assign defaults.
    input_list = [_air_temp, _mrt_, _wind_vel_, _rel_humid]
    input, data_colls = extract_collections(input_list)
    
    if data_colls == []:
        # The inputs are all individual values.
        utci = universal_thermal_climate_index(
            input[0], float(input[1]), input[2], input[3])
        utci_par = UTCIParameter()
        comfort = utci_par.is_comfortable(utci)
        condition = utci_par.thermal_condition(utci)
        category = utci_par.thermal_condition_eleven_point(utci)
    else:
        # The inputs include Data Collections.
        if not isinstance(_air_temp, BaseCollection):
            _air_temp = data_colls[0].get_aligned_collection(
                float(_air_temp), Temperature(), 'C')
        
        comf_obj = UTCI(_air_temp, _rel_humid, _mrt_, _wind_vel_)
        utci = comf_obj.universal_thermal_climate_index
        comfort = comf_obj.is_comfortable
        condition = comf_obj.thermal_condition
        category = comf_obj.thermal_condition_eleven_point


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2020, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Automatically download a .zip file from a URL where climate data resides,
unzip the file, and open .epw, .stat, and ddy weather files.
-

    Args:
        _weather_URL: Text representing the URL at which the climate data resides. 
            To open the a map interface for all publicly availabe climate data (epwmap),
            use the "EPWmap" component.
        _folder_: An optional file path to a directory into which the weather file
            will be downloaded and unziped.  If None, the weather files will be
            downloaded to the ladybug default weather data folder and placed in
            a sub-folder with the name of the weather file location.

    Returns:
        epw_file: The file path of the downloaded epw file.
        stat_file: The file path of the downloaded stat file.
        ddy_file: The file path of the downloaded ddy file.
"""

ghenv.Component.Name = 'LB Download Weather'
ghenv.Component.NickName = 'DownloadEPW'
ghenv.Component.Message = '0.2.3'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '0 :: Import'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from ladybug.futil import unzip_file
    from ladybug.config import folders
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.download import download_file
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

import os


if all_required_inputs(ghenv.Component):
    # name for the weather files
    if _weather_URL.lower().endswith('.zip'):
        # onebuilding URL type
        _folder_name = _weather_URL.split('/')[-1][:-4]
    else:
        # dept of energy URL type
        _folder_name = _weather_URL.split('/')[-2]
    
    # create default working_dir
    if _folder_ is None:
        _folder_ = folders.default_epw_folder
    try:
        _folder_.decode('ascii')
    except UnicodeDecodeError:
        raise UnicodeDecodeError(
            '\nYour download folder "{}" contains non-ASCII characters\n'
            '\nUse the _folder_ input to this component to download EPW files'
            ' to a valid location.'.format(_folder_))
    else:
        print 'Files will be downloaded to: {}'.format(_folder_)
    
    
    # default file names
    epw = os.path.join(_folder_, _folder_name, _folder_name + '.epw')
    stat = os.path.join(_folder_, _folder_name, _folder_name + '.stat')
    ddy = os.path.join(_folder_, _folder_name, _folder_name + '.ddy')
    
    # download and unzip the files if they do not exist
    if not os.path.isfile(epw) or not os.path.isfile(stat) or not os.path.isfile(ddy):
        zip_file_path = os.path.join(_folder_, _folder_name, _folder_name + '.zip')
        download_file(_weather_URL, zip_file_path, True)
        unzip_file(zip_file_path)
    
    # set output
    epw_file, stat_file, ddy_file = epw, stat, ddy


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2020, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Calculate parameters for the relationship between human geometry and the sky given
the position of a human subject and context geometry surrounding this position.
_
The outputs of this component can be plugged into either the "LB Outdoor Solar MRT"
or the "LB Indoor Solar MRT" in order to account for context shading around a
human subject in these MRT calculations.
-

    Args:
        north_: A number between -360 and 360 for the counterclockwise
            difference between the North and the positive Y-axis in degrees.
            90 is West and 270 is East. This can also be Vector for the
            direction to North. (Default: 0)
        _location: A ladybug Location that has been output from the "LB Import EPW"
            component, the "LB Import Location" component, or the "LB Construct
            Location" component. This will be used to compute hourly sun
            positions for the fract_body_exp.
        _position: A point for the position of the human subject in the Rhino scene.
            This is used to understand where a person is in relationship to the
            _context. The point input here should be at the feet of the human
            a series of points will be generated above. This can also be a list
            of points, which will result in several outputs.
        _context: Rhino Breps and/or Rhino Meshes representing context geometry
            that can block the human subject's direct sun and view to the sky.
        _pt_count_: A positive integer for the number of points used to represent
            the human subject geometry. Points are evenly distributed over the
            _height_ and are used to compute fracitonal values for the
            fract_body_exp in the case that only some of the points can see the
            sun. When context shade around the subject is large or coarse,
            using a single point is likely to return similar results as using
            several points. However, this number should be increased when
            context is detailed and has the potential to shade only part of
            the human subject at a given time. (Default: 1).
        _height_: A number for the the height of the human subject in the current Rhino
            Model units. (Default: 1.8 m in the equivalent Rhino Model units;
            roughly the average height of a standing adult).
        parallel_: Set to "True" to run the study using multiple CPUs. This can
            dramatically decrease calculation time but can interfere with
            other computational processes that might be running on your
            machine. (Default: False).
        _run: Set to "True" to run the component and compute the human/sky relationship.
            If set to "False" but all other required inputs are specified, this
            component will output points showing the human subject.

    Returns:
        report: ...
        human_points: The points used to represent the human subject in the calculation
            of the fraction of the body exposed to sun. Note that these are
            generated even when _run is set to "False".
        human_line: Line representing the height of the human subject. Note that this
            is generated even when _run is set to "False".
        fract_body_exp: A data collection for the fraction of the body exposed to
            direct sunlight at each hour of the year. This can be plugged into
            the "Solar MRT" components in order to account for context shading
            in the computation of MRT.
        sky_exposure: A single number between 0 and 1 for the fraction of the sky
            vault in human subject’s view. This can be plugged into the
            "Solar MRT" components in order to account for context shading
            in the computation of MRT.
"""

ghenv.Component.Name = "LB Human to Sky Relation"
ghenv.Component.NickName = 'HumanToSky'
ghenv.Component.Message = '0.1.0'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '3 :: Analyze Geometry'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
    from ladybug_geometry.geometry3d.line import LineSegment3D
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_geometry:\n\t{}'.format(e))

try:
    from ladybug.sunpath import Sunpath
    from ladybug.viewsphere import view_sphere
    from ladybug.datacollection import HourlyContinuousCollection
    from ladybug.header import Header
    from ladybug.analysisperiod import AnalysisPeriod
    from ladybug.datatype.fraction import Fraction
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import conversion_to_meters
    from ladybug_rhino.togeometry import to_point3d, to_vector3d
    from ladybug_rhino.fromgeometry import from_point3d, from_vector3d, \
        from_linesegment3d
    from ladybug_rhino.intersect import join_geometry_to_mesh, intersect_mesh_rays
    from ladybug_rhino.grasshopper import all_required_inputs, list_to_data_tree
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def human_height_points(position, height, pt_count):
    """Get a list of points and a line representing the human geometry.

    Args:
        position: Rhino point for the position of the human.
        height: Number for the height of the human.
        pt_count: Integer for the number of points representing the human.

    Returns:
         A tuple with human points as first element and human line as second.
         Both geomtries are Rhino geometries.
    """
    lb_feet_pt = to_point3d(position).move(Vector3D(0, 0, height / 100))
    lb_hum_line = LineSegment3D(lb_feet_pt, Vector3D(0, 0, height))
    lb_pts = [lb_hum_line.midpoint] if pt_count == 1 else \
        lb_hum_line.subdivide_evenly(pt_count - 1)
    if len(lb_pts) == pt_count - 1:  # sometimes tolerance kills the last point
        lb_pts.append(lb_feet_pt.move(Vector3D(0, 0, height)))
    h_points = [from_point3d(pt) for pt in lb_pts]
    return h_points, from_linesegment3d(lb_hum_line)


def fract_exposed_from_mtx(person_sun_int_matrix, day_pattern):
    """Get a Data Collection of fraction exposed values from an intersection matrix.

    Args:
        person_sun_int_matrix: An intersection matrix of 0s and 1s for the points
            of a single person.
        day_pattern: A list of 8760 booleans indicating whether the sun is
            up (True) or down (Fasle).

    Returns:
         A data collection for the fraction of body exposed.
     """
    pt_count = len(person_sun_int_matrix)
    fract_per_sun = [sum(pt_int_ar) / pt_count for pt_int_ar in zip(*person_sun_int_matrix)]
    fract_exp_vals = []
    per_sun_i = 0
    for is_sun in day_pattern:
        if is_sun:
            fract_exp_vals.append(fract_per_sun[per_sun_i])
            per_sun_i += 1
        else:
            fract_exp_vals.append(0)
    meta_dat = {'type': 'Fraction of Body Exposed to Direct Sun'}
    fract_exp_head = Header(Fraction(), 'fraction', AnalysisPeriod(), meta_dat)
    return HourlyContinuousCollection(fract_exp_head, fract_exp_vals)


def sky_exposure_from_mtx(person_sky_int_matrix, patch_weights):
    """Get a the sky exposure from an intersection matrix.

    Args:
        person_sky_int_matrix: An intersection matrix of 0s and 1s for the
            points of a person intersected with the 145 tregenza patches.
        patch_weights: A list of 145 weights to be applies to the patches.

    Returns:
         A value for the sky exposure of the person.
     """
    pt_count = len(person_sky_int_matrix)
    sky_exp_per_pt = [sum((r * w) / 145 for r, w in zip(int_list, patch_wghts))
                      for int_list in person_sky_int_matrix]
    return sum(sky_exp_per_pt) / pt_count


if all_required_inputs(ghenv.Component):
    # process the north input if specified
    if north_ is not None:  # process the north_
        try:
            north_ = math.degrees(to_vector2d(north_).angle_clockwise(Vector2D(0, 1)))
        except AttributeError:  # north angle instead of vector
            north_ = float(north_)
    else:
        north_ = 0

    # set the default point count and height if unspecified
    _pt_count_ = _pt_count_ if _pt_count_ is not None else 1
    _height_ = _height_ if _height_ is not None else 1.8 / conversion_to_meters()

    # create the points representing the human geometry
    human_points = []
    human_line = []
    for pos in _position:
        hpts, hlin = human_height_points(pos, _height_, _pt_count_)
        human_points.extend(hpts)
        human_line.append(hlin)

    if _run:
        # mesh the context for the intersection calculation
        shade_mesh = join_geometry_to_mesh(_context)

        # generate the sun vectors for each sun-up hour of the year
        sp = Sunpath.from_location(_location, north_)
        sun_vecs = []
        day_pattern = []
        for hoy in range(8760):
            sun = sp.calculate_sun_from_hoy(hoy)
            day_pattern.append(sun.is_during_day)
            if sun.is_during_day:
                sun_vecs.append(from_vector3d(sun.sun_vector_reversed))

        # intersect the sun vectors with the context and compute fraction exposed
        sun_int_matrix, angles = intersect_mesh_rays(
            shade_mesh, human_points, sun_vecs, parallel=parallel_)
        fract_body_exp = []
        for i in range(0, len(human_points), _pt_count_):
            fract_body_exp.append(
                fract_exposed_from_mtx(sun_int_matrix[i:i + _pt_count_], day_pattern))

        # generate the vectors and weights for sky exposure
        sky_vecs = [from_vector3d(vec) for vec in view_sphere.tregenza_dome_vectors]
        patch_wghts = view_sphere.dome_patch_weights(1)

        # compute the sky exposure
        sky_int_matrix, angles = intersect_mesh_rays(
            shade_mesh, human_points, sky_vecs, parallel=parallel_)
        sky_exposure = []
        for i in range(0, len(human_points), _pt_count_):
            sky_exposure.append(
                sky_exposure_from_mtx(sky_int_matrix[i:i + _pt_count_], patch_wghts))



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2020, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Calculate Predicted Mean Vote (PMV).
-
PMV is a thermal comfort model for use on the interior of buildings
where a heating or cooling system is operational.
Note that, for naturally ventilated buildings, the Adaptive thermal
comfort model should be used.
-

    Args:
        _out_temp: Outdoor temperatures in one of the following formats:
            _
            1 - A Data Collection of prevailing outdoor temperature values in C.
                This Data Collection must align with the _air_temp or _mrt_
                inputs and bear the PrevailingOutdoorTemperature data type in
                its header.
            2 - A single prevailing outdoor temperature value in C to be used
                for all of the _air_temp or _mrt_ inputs.
            3 - A Data Collection of outdoor dry bulb temperatures recorded over
                the entire year. This Data Collection must be continouous and
                must either be an Hourly Collection or Daily Collection. In the event
                that the input adapt_par_ has a _avgm_or_runmean_ set to True,
                Monthly collections are also acceptable here. Note that, because
                an annual input is required, this input collection does not have
                to align with the _air_temp or _mrt_ inputs.
        _air_temp: Data Collection or individual value of air temperature in C.
        _mrt_: Data Collection or individual value of mean radiant temperature
            (MRT) in C. Default is the same as the air_temp.
        _air_speed_: Data Collection or individual of air speed value in m/s.
            Default is a very low speed of 0.1 m/s.
        adapt_par_: Optional Adaptive Parameter object to specify parameters 
            under which conditions are considered acceptable. If None, default
            will assume ASHRAE-55 criteria.
        _run: Set to True to run the component.

    Returns:
        report: Reports, errors, warnings, etc.
        prevail_temp: Data Collection of prevailing outdoor temperature in
            degrees C.
        neutral_temp: Data Collection of the desired neutral temperature in
            degrees C.
        deg_neutral: Data Collection of the degrees from desired neutral
            temperature in degrees C.
        comfort: Integers noting whether the input conditions are acceptable
            according to the assigned comfort_parameter.
            _
            Values are one of the following:
                0 = uncomfortable
                1 = comfortable
        condition: Integers noting the thermal status of a subject according to
            the assigned comfort_parameter.
            _
            Values are one of the following:
                -1 = cold
                 0 = netural
                +1 = hot
        comf_obj: A Python object containing all inputs and results of the
            analysis.  This can be plugged into components like the "Comfort
            Statistics" component to get further information.
"""

ghenv.Component.Name = 'LB Adaptive Comfort'
ghenv.Component.NickName = 'Adaptive'
ghenv.Component.Message = '0.1.3'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '5'

try:
    from ladybug.datatype.temperature import Temperature
    from ladybug.datacollection import BaseCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_comfort.collection.adaptive import Adaptive
    from ladybug_comfort.parameter.adaptive import AdaptiveParameter
    from ladybug_comfort.adaptive import t_operative, \
        adaptive_comfort_ashrae55, adaptive_comfort_en15251, \
        cooling_effect_ashrae55, cooling_effect_en15251, \
        adaptive_comfort_conditioned
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_comfort:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


def extract_collections(input_list):
    """Process inputs into collections and floats."""
    defaults = [None, None, _air_temp, 0.1]
    data_colls = []
    for i, input in enumerate(input_list):
        if input is None:
            input_list[i] = defaults[i]
        elif isinstance(input, BaseCollection):
            data_colls.append(input)
        else:
            try:
                input_list[i] = float(input)
            except ValueError as e:
                raise TypeError('input {} is not valid. Expected float or '
                                'DataCollection. Got {}'.format(input, type(input)))
    return input_list, data_colls

if all_required_inputs(ghenv.Component) and _run is True:
    # Process inputs and assign defaults.
    input_list = [_out_temp, _air_temp, _mrt_, _air_speed_]
    input, data_colls = extract_collections(input_list)
    adapt_par = adapt_par_ or AdaptiveParameter()

    if data_colls == []:
        # The inputs are all individual values.
        prevail_temp = input[0]
        to = t_operative(input[1], float(input[2]))
        
        # Determine the ralationship to the neutral temperature
        if adapt_par.conditioning != 0:
            comf_result = adaptive_comfort_conditioned(prevail_temp, to,
                adapt_par.conditioning, adapt_par.standard)
        elif adapt_par.ashrae55_or_en15251 is True:
            comf_result = adaptive_comfort_ashrae55(prevail_temp, to)
        else:
            comf_result = adaptive_comfort_en15251(prevail_temp, to)
        
        # Determine the cooling effect
        if adapt_par.discrete_or_continuous_air_speed is True:
            ce = cooling_effect_ashrae55(input[3], to)
        else:
            ce = cooling_effect_en15251(input[3], to)
        
        # Output results
        neutral_temp = comf_result['t_comf']
        deg_neutral = comf_result['deg_comf']
        comfort = adapt_par.is_comfortable(comf_result, ce)
        condition = adapt_par.thermal_condition(comf_result, ce)
    else:
        # The inputs include Data Collections.
        if not isinstance(_air_temp, BaseCollection):
            _air_temp = data_colls[0].get_aligned_collection(
                float(_air_temp), Temperature(), 'C')
        
        comf_obj = Adaptive.from_air_and_rad_temp(_out_temp, _air_temp, _mrt_,
                                                 _air_speed_, adapt_par)
        prevail_temp = comf_obj.prevailing_outdoor_temperature
        neutral_temp = comf_obj.neutral_temperature
        deg_neutral = comf_obj.degrees_from_neutral
        comfort = comf_obj.is_comfortable
        condition = comf_obj.thermal_condition


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2020, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Deconstruct a Ladybug Header into its components.
-

    Args:
        _header: The header of the DataCollection (containing metadata).

    Returns:
        data_type: Type of data (e.g. Temperature) (Default: unknown).
        unit: Units of the data_type (e.g. C) (Default: unknown)
        a_period: A Ladybug AnalysisPeriod object.
        metadata: Optional metadata associated with the Header.
"""

ghenv.Component.Name = "LB Deconstruct Header"
ghenv.Component.NickName = 'XHeader'
ghenv.Component.Message = '0.1.3'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '0'

try:
    from ladybug.header import Header
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    assert isinstance(_header, Header), 'Expected Ladybug Header. ' \
        'Got {}.'.format(type(_header))
    data_type = _header.data_type
    unit = _header.unit
    a_period = _header.analysis_period
    metadata = [': '.join([key, val]) for key, val in _header.metadata.items()]


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2020, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Calculate the incident radiation on geometry using a sky matrix from the "Cumulative
Sky Matrix" component.
_
Such studies of incident radiation can be used to apprxomiate the energy that can
be collected from photovoltaic or solar thermal systems. They are also useful
for evaluating the impact of a building's windows on both energy use and the
size/cost of cooling systems. For studies of cooling system size/cost, a sky
matrix derived from the STAT file's clear sky radiation should be used. For
studies of energy use impact, such as the evaluation of passive solar heating
or the potential for excessive cooling energy use, a matrix from EPW radiation
should be used.
_
Note that NO REFLECTIONS OF SOLAR ENERGY ARE INCLUDED IN THE ANALYSIS
PERFORMED BY THIS COMPONENT and it is important to bear in mind that vertical
surfaces typically receive 20% - 30% of their solar energy from reflection off
of the ground. Also note that this component uses the CAD environment's ray
intersection methods, which can be fast for geometries with low complexity
but does not scale well for complex geometries or many test points. For such
complex cases and situations where relfection of solar energy are important,
honeybee-radiance should be used.
-

    Args:
        _sky_mtx: A Sky Matrix from the "LB Cumulative Sky Matrix" component, which
            describes the radiation coming from the various patches of the sky.
            The "LB Sky Dome" component can be used to visualize any sky matrix
            to understand its relationship to the test geometry.
        _geometry: Rhino Breps and/or Rhino Meshes for which incident radiation analysis
            will be conducted. If Breps are input, they will be subdivided using
            the _grid_size to yeild individual points at which analysis will
            occur. If a Mesh is input, radiation analysis analysis will be
            performed for each face of this mesh instead of subdividing it.
        context_: Rhino Breps and/or Rhino Meshes representing context geometry
            that can block solar radiation to the test _geometry.
        _grid_size: A positive number in Rhino model units for the size of grid
            cells at which the input _geometry will be subdivided for incident
            radiation analysis. The smaller the grid size, the higher the
            resolution of the analysis and the longer the calculation will take.
            So it is recommended that one start with a large value here and
            decrease the value as needed. However, the grid size should usually
            be smaller than the dimensions of the smallest piece of the _geometry
            and context_ in order to yield meaningful results.
        _offset_dist_: A number for the distance to move points from the surfaces
            of the input _geometry.  Typically, this should be a small positive
            number to ensure points are not blocked by the mesh. (Default: 10 cm
            in the equivalent Rhino Model units).
        legend_par_: Optional legend parameters from the "LB Legend Parameters"
            that will be used to customize the display of the results.
        parallel_: Set to "True" to run the study using multiple CPUs. This can
            dramatically decrease calculation time but can interfere with
            other computational processes that might be running on your
            machine. (Default: False).
        _run: Set to "True" to run the component and perform incident radiation
            analysis. If set to "False" but all other required inputs are
            specified, this component will output points showing the resolution
            of the analysis (but not run the study) so that an estimate of study
            run time can be obtained.

    Returns:
        report: ...
        points: The grid of points on the test _geometry that are be used to perform
            the incident radiation analysis.  Note that these points are generated
            even when _run is set to "False" so that an estimate of study run
            time can be obtained.
        results: A list of numbers that aligns with the points. Each number indicates
            the cumulative incident radiation received by each of the points
            from the sky matrix in kWh/m2.
        total: A number for the total incident solar energy falling on all input geometry
            in kWh. Note that, unlike the radiation results above, which are
            normlaized by area, these values are not area-normalized and so
            the input geometry must be represented correctly in the Rhino
            model's unit system in order for this output to be meaningful.
        mesh: A colored mesh of the test _geometry representing the cumulative
            incident radiation received by the input _geometry.
        legend: A legend showing the kWh/m2 that correspond to the colors of the mesh.
        title: A text object for the study title.
        int_mtx: A Matrix object that can be connected to the "LB Deconstruct Matrix"
            component to obtain detailed patch-by-patch results of the study.
            Each sub-list of the matrix (aka. branch of the Data Tree) represents
            one of the points used for analysis. The length of each sub-list
            matches the number of sky patches in the input sky matrix (145 for
            the default Tregenza sky and 577 for the high_density Reinhart sky).
            Each value in the sub-list is a value between 0 and 1 indicating the
            relationship between the point and the patch of the sky. A value of
            "0", indicates that the patch is not visible for that point at all
            while a value of "1" indicates that the patch hits the surface that
            the point represents head on.
"""

ghenv.Component.Name = "LB Incident Radiation"
ghenv.Component.NickName = 'IncidentRadiation'
ghenv.Component.Message = '0.1.2'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '3 :: Analyze Geometry'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

import math
try:  # python 2
    from itertools import izip as zip
except ImportError:  # python 3
    pass

try:
    from ladybug.viewsphere import view_sphere
    from ladybug.graphic import GraphicContainer
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import conversion_to_meters
    from ladybug_rhino.togeometry import to_joined_gridded_mesh3d
    from ladybug_rhino.fromgeometry import from_mesh3d, from_point3d, from_vector3d
    from ladybug_rhino.fromobjects import legend_objects
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.intersect import join_geometry_to_mesh, intersect_mesh_rays
    from ladybug_rhino.grasshopper import all_required_inputs, hide_output, \
        show_output, objectify_output, de_objectify_output
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # set the default offset distance
    _offset_dist_ = _offset_dist_ if _offset_dist_ is not None \
        else 0.1 / conversion_to_meters()

    # create the gridded mesh from the geometry
    study_mesh = to_joined_gridded_mesh3d(_geometry, _grid_size, _offset_dist_)
    points = [from_point3d(pt) for pt in study_mesh.face_centroids]

    if _run:  # run the entire study
        # hide the study points and mesh the geometry and context
        hide_output(ghenv.Component, 1)
        shade_mesh = join_geometry_to_mesh(_geometry + context_)

        # deconstruct the matrix and get the sky dome vectors
        mtx = de_objectify_output(_sky_mtx)
        total_sky_rad = [dir_rad + dif_rad for dir_rad, dif_rad in zip(mtx[1], mtx[2])]
        lb_vecs = view_sphere.tregenza_dome_vectors if len(total_sky_rad) == 145 \
            else view_sphere.reinhart_dome_vectors
        if mtx[0][0] != 0:  # there is a north input for sky; rotate vectors
            north_angle = math.radians(mtx[0][0])
            lb_vecs = [vec.rotate_xy(north_angle) for vec in lb_vecs]
        sky_vecs = [from_vector3d(vec) for vec in lb_vecs]

        # intersect the rays with the mesh
        normals = [from_vector3d(vec) for vec in study_mesh.face_normals]
        int_matrix_init, angles = intersect_mesh_rays(
            shade_mesh, points, sky_vecs, normals, parallel=parallel_)

        # compute the results
        results = []
        int_matrix = []
        for int_vals, angles in zip(int_matrix_init, angles):
            pt_rel = [ival * math.cos(ang) for ival, ang in zip(int_vals, angles)]
            int_matrix.append(pt_rel)
            rad_result = sum(r * w for r, w in zip(pt_rel, total_sky_rad))
            results.append(rad_result)

        # output the intersection matrix and compute total radiation
        int_mtx = objectify_output('Geometry/Sky Intersection Matrix', int_matrix)
        unit_conv = conversion_to_meters() ** 2
        total = 0
        for rad, area in zip(results, study_mesh.face_areas):
            total += rad * area * unit_conv

        # create the mesh and legend outputs
        graphic = GraphicContainer(results, study_mesh.min, study_mesh.max, legend_par_)
        graphic.legend_parameters.title = 'kWh/m2'
        title = text_objects(
            'Incident Radiation', graphic.lower_title_location,
            graphic.legend_parameters.text_height * 1.5,
            graphic.legend_parameters.font)

        # create all of the visual outputs
        study_mesh.colors = graphic.value_colors
        mesh = from_mesh3d(study_mesh)
        legend = legend_objects(graphic.legend)

    else:  # only show the points
        show_output(ghenv.Component, 1)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2020, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Calculate the number of hours of direct sunlight received by geometry using sun
vectors obtained from the "LB SunPath" component.
_
Such direct sun calculations can be used for shadow studies of outdoor enviroments
or can be used to estimate glare potential from direct sun on the indoors.
_
Note that this component uses the CAD environment's ray intersection methods,
which can be fast for geometries with low complexity but does not scale well
for complex geometries or many test points. For such complex studies,
honeybee-radiance should be used.
-

    Args:
        _vectors: Sun vectors from the "LB SunPath" component, which will be used
            to determine the number of hours of direct sunlight received by the
            test _geometry.
        _timestep_: A positive integer for the number of timesteps per hour at
            which the "LB SunPath" component generated sun vectors. This is used
            to correctly interpret the time duration represented by each of
            the input sun vectors. (Default: 1 for 1 vector per hour).
        _geometry: Rhino Breps and/or Rhino Meshes for which direct sun analysis
            will be conducted. If Breps are input, they will be subdivided using
            the _grid_size to yeild individual points at which analysis will
            occur. If a Mesh is input, direct sun analysis analysis will be
            performed for each face of this mesh instead of subdividing it.
        context_: Rhino Breps and/or Rhino Meshes representing context geometry
            that can block sunlight to the test _geometry.
        _grid_size: A positive number in Rhino model units for the size of grid
            cells at which the input _geometry will be subdivided for direct sun
            analysis. The smaller the grid size, the higher the resolution of
            the analysis and the longer the calculation will take.  So it is
            recommended that one start with a large value here and decrease
            the value as needed. However, the grid size should usually be
            smaller than the dimensions of the smallest piece of the _geometry
            and context_ in order to yield meaningful results.
        _offset_dist_: A number for the distance to move points from the surfaces
            of the input _geometry.  Typically, this should be a small positive
            number to ensure points are not blocked by the mesh. (Default: 10 cm
            in the equivalent Rhino Model units).
        legend_par_: Optional legend parameters from the "LB Legend Parameters"
            that will be used to customize the display of the results.
        parallel_: Set to "True" to run the study using multiple CPUs. This can
            dramatically decrease calculation time but can interfere with
            other computational processes that might be running on your
            machine. (Default: False).
        _run: Set to "True" to run the component and perform direct sun analysis.
            If set to "False" but all other required inputs are specified, this
            component will output points showing the resolution of the analysis
            (but not run the study) so that an estimate of study run time can
            be obtained.

    Returns:
        report: ...
        points: The grid of points on the test _geometry that are be used to perform
            the direct sun analysis.  Note that these points are generated even
            when _run is set to "False" so that an estimate of study run time can
            be obtained.
        results: A list of numbers that aligns with the points. Each number indicates
            the number of hours of direct sunlight received by each of the
            points.  Note that is is the number of hours out of the total
            number of connected _vectors.
        mesh: A colored mesh of the test _geometry representing the hours of direct
            sunlight received by this input _geometry
        legend: A legend showing the number of hours that correspond to the colors
            of the mesh.
        title: A text object for the study title.
        int_mtx: A Matrix object that can be connected to the "LB Deconstruct Matrix"
            component to obtain detailed vector-by-vector results of the study.
            Each sub-list of the matrix (aka. branch of the Data Tree) represents
            one of the points used for analysis. The length of each sub-list
            matches the number of _vectors used for the analysis. Each value
            in the sub-list is either a "1", indicating that the sun is visible
            for that vector, or a "0", indicating that the sun is not visible
            for that vector.
"""

ghenv.Component.Name = "LB Direct Sun Hours"
ghenv.Component.NickName = 'DirectSunHours'
ghenv.Component.Message = '0.1.1'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '3 :: Analyze Geometry'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from ladybug.color import Colorset
    from ladybug.graphic import GraphicContainer
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.config import conversion_to_meters
    from ladybug_rhino.togeometry import to_joined_gridded_mesh3d, to_vector3d
    from ladybug_rhino.fromgeometry import from_mesh3d, from_point3d, from_vector3d
    from ladybug_rhino.fromobjects import legend_objects
    from ladybug_rhino.text import text_objects
    from ladybug_rhino.intersect import join_geometry_to_mesh, intersect_mesh_rays
    from ladybug_rhino.grasshopper import all_required_inputs, hide_output, \
        show_output, objectify_output
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # set the default offset distance
    _offset_dist_ = _offset_dist_ if _offset_dist_ is not None \
        else 0.1 / conversion_to_meters()

    # create the gridded mesh from the geometry
    study_mesh = to_joined_gridded_mesh3d(_geometry, _grid_size, _offset_dist_)
    points = [from_point3d(pt) for pt in study_mesh.face_centroids]

    if _run:  # run the entire study
        # hide the study points and mesh the geometry and context
        hide_output(ghenv.Component, 1)
        shade_mesh = join_geometry_to_mesh(_geometry + context_)

        # get the study points and reverse the sun vectors (for backward ray-tracting)
        rev_vec = [from_vector3d(to_vector3d(vec).reverse()) for vec in _vectors]
        normals = [from_vector3d(vec) for vec in study_mesh.face_normals]

        # intersect the rays with the mesh
        int_matrix, angles = intersect_mesh_rays(
            shade_mesh, points, rev_vec, normals, parallel=parallel_)

        # compute the results
        int_mtx = objectify_output('Sun Intersection Matrix', int_matrix)
        if _timestep_ and _timestep_ != 1:  # divide by the timestep before output
            results = [sum(int_list) / _timestep_ for int_list in int_matrix]
        else:  # no division required
            results = [sum(int_list) for int_list in int_matrix]

        # create the mesh and legend outputs
        graphic = GraphicContainer(results, study_mesh.min, study_mesh.max, legend_par_)
        graphic.legend_parameters.title = 'hours'
        if legend_par_ is None or legend_par_.are_colors_default:
            graphic.legend_parameters.colors = Colorset.ecotect()
        title = text_objects('Direct Sun Hours', graphic.lower_title_location,
                             graphic.legend_parameters.text_height * 1.5,
                             graphic.legend_parameters.font)

        # create all of the visual outputs
        study_mesh.colors = graphic.value_colors
        mesh = from_mesh3d(study_mesh)
        legend = legend_objects(graphic.legend)

    else:  # only show the the test points
        show_output(ghenv.Component, 1)



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2020, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Construct a Ladybug data collection from header and values.
-

    Args:
        _header:A Ladybug header object describing the metadata of the data collection.
        _values: A list of numerical values for the data collection.

    Returns:
        data: A Ladybug data collection object.
"""

ghenv.Component.Name = "LB Construct Data"
ghenv.Component.NickName = '+Data'
ghenv.Component.Message = '0.1.2'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '1 :: Analyze Data'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from ladybug.datacollection import HourlyContinuousCollection
except ImportError as e:
    raise ImportError('\nFailed to import ladybug:\n\t{}'.format(e))

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    data = HourlyContinuousCollection(_header, _values)


# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2020, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Open a file's directory in Windows Explorer or Mac Finder.
This is useful for understanding weather data or simulation files.
-

    Args:
        _file_path: Full path to a file or directory to be opened in Explorer/Finder.
    
    Returns:
        report: Reports, errors, warnings, etc.
"""

ghenv.Component.Name = 'LB Open Directory'
ghenv.Component.NickName = 'OpenDir'
ghenv.Component.Message = '0.1.1'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '4 :: Extra'
ghenv.Component.AdditionalHelpFromDocStrings = '5'

import os
import subprocess

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))


if all_required_inputs(ghenv.Component):
    # check that the file exists
    assert os.path.exists(_file_path), \
        'No file or directory was found at: {}'.format(_file_path)
    folder = _file_path if os.path.isdir(_file_path) else os.path.dirname(_file_path)

    # open the file in explorer or finder
    if os.name == 'nt':  # we are on Windows
        subprocess.Popen('explorer.exe ' + folder)
    else:  # assume we are on Mac
        subprocess.call(['open', '-R', _file_path])



# Ladybug: A Plugin for Environmental Analysis (GPL)
# This file is part of Ladybug.
#
# Copyright (c) 2020, Ladybug Tools.
# You should have received a copy of the GNU General Public License
# along with Ladybug; If not, see <http://www.gnu.org/licenses/>.
# 
# @license GPL-3.0+ <http://spdx.org/licenses/GPL-3.0+>

"""
Open EPWmap in a web browser.
-

    Args:
        _epw_map: Set to "True" to open EPWmap in a supported browser.
    Returns:
        report: ...
"""

ghenv.Component.Name = 'LB EPWmap'
ghenv.Component.NickName = 'EPWMap'
ghenv.Component.Message = '0.1.1'
ghenv.Component.Category = 'Ladybug'
ghenv.Component.SubCategory = '0 :: Import'
ghenv.Component.AdditionalHelpFromDocStrings = '1'

try:
    from ladybug_rhino.grasshopper import all_required_inputs
except ImportError as e:
    raise ImportError('\nFailed to import ladybug_rhino:\n\t{}'.format(e))

import webbrowser as wb
import os


# dictonary of accetable browsers and their default file paths.
acceptable_browsers = [
    ['chrome', 'C:\\Program Files (x86)\\Google\\Chrome\\Application\\chrome.exe'],
    ['firefox', 'C:\\Program Files\\Mozilla Firefox\\firefox.exe'],
    ['chrome', '/Applications/Google Chrome.app/Contents/MacOS/Google Chrome'] # MacOS
    ]

# URL to epwmap.
url = 'http://www.ladybug.tools/epwmap/'

# function for opening a browser page on Mac
def mac_open(url):
    os.system("open \"\" " + url)

if all_required_inputs(ghenv.Component) and _epw_map is True:
    broswer_found = False
    for browser in acceptable_browsers:
        browser_path = browser[1]
        if broswer_found == False and os.path.isfile(browser_path) == True:
            broswer_found = True
            wb.register(browser[0],  None, wb.BackgroundBrowser(browser_path), 1)
            try:
                wb.get(browser[0]).open(url, 2, True)
                print "Opening epwmap."
            except ValueError:
                mac_open(url)
    if broswer_found == False:
       print "An accepable broswer was not found on your machine.\n" \
       "The default browser will be used but epwmap may not display correctly there."
       try:
        wb.open(url, 2, True)
       except ValueError:
        mac_open(url)
else:
    print "Set _epw_map to True."